"""Fetching repositories from URLs."""
import enum
import pathlib
from collections.abc import Callable
from typing import Optional

from yarl import URL

from cutty.repositories.domain.matchers import Matcher
from cutty.repositories.domain.revisions import Revision
from cutty.repositories.domain.stores import defaultstore
from cutty.repositories.domain.stores import Store


class FetchMode(enum.Enum):
    """Fetch mode."""

    ALWAYS = enum.auto()
    AUTO = enum.auto()
    NEVER = enum.auto()


Fetcher = Callable[[URL, Store, Optional[Revision], FetchMode], Optional[pathlib.Path]]
FetchFunction = Callable[[URL, pathlib.Path, Optional[Revision]], None]
FetchDecorator = Callable[[FetchFunction], Fetcher]


def fetcher(*, match: Matcher, store: Store = defaultstore) -> FetchDecorator:
    """A fetcher retrieves a repository from a URL into storage."""
    relativestore = store

    def _decorator(fetch: FetchFunction) -> Fetcher:
        def _fetcher(
            url: URL, store: Store, revision: Optional[Revision], mode: FetchMode
        ) -> Optional[pathlib.Path]:
            if not match(url):
                return None

            destination = store(url) / relativestore(url)

            if (
                mode is FetchMode.ALWAYS
                or mode is FetchMode.AUTO
                and not destination.exists()
            ):
                fetch(url, destination, revision)

            return destination

        return _fetcher

    return _decorator



"""Configuration."""
from dataclasses import dataclass
from typing import Any

from cutty.templates.domain.variables import Variable


@dataclass
class Config:
    """Template configuration."""

    settings: dict[str, Any]
    variables: tuple[Variable, ...]



"""Create a project from a Cookiecutter template."""
import pathlib
from collections.abc import Mapping
from types import MappingProxyType
from typing import Optional

import appdirs
from lazysequence import lazysequence

from cutty.filestorage.adapters.cookiecutter import createcookiecutterstorage
from cutty.filesystems.domain.purepath import PurePath
from cutty.repositories.adapters.storage import getdefaultrepositoryprovider
from cutty.templates.adapters.cookiecutter.binders import bindcookiecuttervariables
from cutty.templates.adapters.cookiecutter.config import findhooks
from cutty.templates.adapters.cookiecutter.config import findpaths
from cutty.templates.adapters.cookiecutter.config import loadcookiecutterconfig
from cutty.templates.adapters.cookiecutter.render import createcookiecutterrenderer
from cutty.templates.domain.bindings import Binding
from cutty.templates.domain.renderfiles import renderfiles


def create(
    template: str,
    *,
    extra_context: Mapping[str, str] = MappingProxyType({}),
    no_input: bool = False,
    checkout: Optional[str] = None,
    outputdir: Optional[pathlib.Path] = None,
    directory: Optional[pathlib.PurePosixPath] = None,
    overwrite_if_exists: bool = False,
    skip_if_file_exists: bool = False,
    outputdirisproject: bool = False,
) -> None:
    """Generate a project from a Cookiecutter template."""
    cachedir = pathlib.Path(appdirs.user_cache_dir("cutty"))
    templatedir = getdefaultrepositoryprovider(cachedir)(template, revision=checkout)

    if directory is not None:
        templatedir = templatedir.joinpath(*directory.parts)  # pragma: no cover

    if outputdir is None:  # pragma: no branch
        outputdir = pathlib.Path.cwd()

    config = loadcookiecutterconfig(template, templatedir)
    render = createcookiecutterrenderer(templatedir, config)
    bindings = bindcookiecuttervariables(
        config.variables,
        render,
        interactive=not no_input,
        bindings=[Binding(key, value) for key, value in extra_context.items()],
    )

    projectfiles = lazysequence(
        renderfiles(findpaths(templatedir, config), render, bindings)
    )
    if not projectfiles:  # pragma: no cover
        return

    if outputdirisproject:
        projectdir = outputdir
    else:
        projectdir = outputdir / projectfiles[0].path.parts[0]

    hookfiles = lazysequence(renderfiles(findhooks(templatedir), render, bindings))

    with createcookiecutterstorage(
        outputdir,
        projectdir,
        overwrite_if_exists,
        skip_if_file_exists,
        hookfiles,
    ) as storage:
        for projectfile in projectfiles.release():
            if outputdirisproject:
                path = PurePath(*projectfile.path.parts[1:])
                projectfile = projectfile.withpath(path)
            storage.add(projectfile)



"""Cookiecutter file storage."""
import pathlib
import platform
import shutil
import subprocess  # noqa: S404
import sys
import tempfile
from collections.abc import Iterable

from cutty.filestorage.adapters.disk import DiskFileStorage
from cutty.filestorage.adapters.disk import FileExistsPolicy
from cutty.filestorage.domain.files import File
from cutty.filestorage.domain.observers import FileStorageObserver


class _Hooks:
    def __init__(self, *, hookfiles: Iterable[File], project: pathlib.Path) -> None:
        self.hooks = {hook.path.stem: hook for hook in hookfiles}
        self.project = project

    def run(self, hook: str) -> None:
        hookfile = self.hooks.get(hook)
        if hookfile is not None:
            self._runfile(hookfile)

    def _runfile(self, hookfile: File) -> None:
        with tempfile.TemporaryDirectory() as root:
            with DiskFileStorage(pathlib.Path(root)) as storage:
                storage.add(hookfile)
                path = storage.resolve(hookfile.path)
                self._runpath(path)

    def _runpath(self, path: pathlib.Path) -> None:
        command = (
            [pathlib.Path(sys.executable), path] if path.suffix == ".py" else [path]
        )
        shell = platform.system() == "Windows"
        self.project.mkdir(parents=True, exist_ok=True)
        subprocess.run(command, shell=shell, cwd=self.project, check=True)  # noqa: S602


class CookiecutterHooksObserver(FileStorageObserver):
    """Storage observer invoking Cookiecutter hooks."""

    def __init__(
        self,
        *,
        hookfiles: Iterable[File],
        project: pathlib.Path,
        fileexists: FileExistsPolicy,
    ) -> None:
        """Initialize."""
        self.hooks = _Hooks(hookfiles=hookfiles, project=project)
        self.fileexists = fileexists

    def begin(self) -> None:
        """A storage transaction was started."""
        self.hooks.run("pre_gen_project")

    def commit(self) -> None:
        """A storage transaction was completed."""
        self.hooks.run("post_gen_project")

    def rollback(self) -> None:
        """A storage transaction was aborted."""
        if self.fileexists is not FileExistsPolicy.OVERWRITE:
            shutil.rmtree(self.hooks.project, ignore_errors=True)



"""Repository providers for ZIP archives."""
from pathlib import Path

from cutty.filesystems.adapters.zip import ZipFilesystem
from cutty.repositories.adapters.fetchers.file import filefetcher
from cutty.repositories.adapters.fetchers.ftp import ftpfetcher
from cutty.repositories.adapters.fetchers.http import httpfetcher
from cutty.repositories.domain.mounters import unversioned_mounter
from cutty.repositories.domain.providers import localprovider
from cutty.repositories.domain.providers import remoteproviderfactory


def match(path: Path) -> bool:
    """Return True if the path is a ZIP archive."""
    return path.suffix.lower() == ".zip" and path.is_file()


mount = unversioned_mounter(ZipFilesystem)
localzipprovider = localprovider(match=match, mount=mount)
zipproviderfactory = remoteproviderfactory(
    match=lambda url: url.path.lower().endswith(".zip"),
    fetch=[httpfetcher, ftpfetcher, filefetcher],
    mount=mount,
)



"""Fetcher for git repositories."""
import contextlib
import pathlib
from typing import Optional

import pygit2
from yarl import URL

from cutty.repositories.domain.fetchers import fetcher
from cutty.repositories.domain.matchers import scheme
from cutty.repositories.domain.revisions import Revision
from cutty.repositories.domain.stores import defaultstore


def _fix_repository_head(repository: pygit2.Repository) -> pygit2.Reference:
    """Work around a bug in libgit2 resulting in a bogus HEAD reference.

    Cloning with a remote callback results in HEAD pointing to the user's
    `init.defaultBranch` instead of the default branch of the cloned repository.
    """
    # https://github.com/libgit2/pygit2/issues/1073
    head = repository.references["HEAD"]

    with contextlib.suppress(KeyError):
        return head.resolve()

    for branch in ["main", "master"]:
        ref = f"refs/heads/{branch}"
        if ref in repository.references:
            head.set_target(ref, message="repair broken HEAD after clone")
            break

    return head.resolve()


@fetcher(
    match=scheme("file", "ftp", "ftps", "git", "http", "https", "ssh"),
    store=lambda url: defaultstore(url).with_suffix(".git"),
)
def gitfetcher(
    url: URL, destination: pathlib.Path, revision: Optional[Revision]
) -> None:
    """Fetch a git repository."""

    def _createremote(
        repository: pygit2.Repository, name: bytes, url: bytes
    ) -> pygit2.Remote:
        name_ = name.decode()
        repository.config[f"remote.{name_}.mirror"] = True
        return repository.remotes.create(name, url, "+refs/*:refs/*")

    if destination.exists():
        repository = pygit2.Repository(destination)
        for remote in repository.remotes:
            remote.fetch(prune=pygit2.GIT_FETCH_PRUNE)
    else:
        _fix_repository_head(
            pygit2.clone_repository(
                str(url), str(destination), bare=True, remote=_createremote
            )
        )



"""File storage for Cookiecutter projects."""
import pathlib
from collections.abc import Sequence

from cutty.filestorage.adapters.disk import DiskFileStorage
from cutty.filestorage.adapters.disk import FileExistsPolicy
from cutty.filestorage.adapters.observers.cookiecutter import CookiecutterHooksObserver
from cutty.filestorage.adapters.observers.git import GitRepositoryObserver
from cutty.filestorage.domain.files import File
from cutty.filestorage.domain.observers import observe
from cutty.filestorage.domain.storage import FileStorage


def fileexistspolicy(
    overwrite_if_exists: bool, skip_if_file_exists: bool
) -> FileExistsPolicy:
    """Return the policy for overwriting existing files."""
    return (
        FileExistsPolicy.RAISE
        if not overwrite_if_exists
        else FileExistsPolicy.SKIP
        if skip_if_file_exists
        else FileExistsPolicy.OVERWRITE
    )


def createcookiecutterstorage(
    outputdir: pathlib.Path,
    project_dir: pathlib.Path,
    overwrite_if_exists: bool,
    skip_if_file_exists: bool,
    hookfiles: Sequence[File],
) -> FileStorage:
    """Create storage for Cookiecutter project files."""
    fileexists = fileexistspolicy(overwrite_if_exists, skip_if_file_exists)
    storage: FileStorage = DiskFileStorage(outputdir, fileexists=fileexists)

    if hookfiles:  # pragma: no branch
        observer = CookiecutterHooksObserver(
            hookfiles=hookfiles, project=project_dir, fileexists=fileexists
        )
        storage = observe(storage, observer)

    storage = observe(storage, GitRepositoryObserver(project=project_dir))

    return storage



"""Rendering files."""
from collections.abc import Iterable
from collections.abc import Iterator
from collections.abc import Sequence

from cutty.filestorage.domain.files import File
from cutty.filestorage.domain.files import loadfile
from cutty.filesystems.domain.path import Path
from cutty.templates.domain.bindings import Binding
from cutty.templates.domain.render import Renderer


def renderfiles(
    paths: Iterable[Path], render: Renderer, bindings: Sequence[Binding]
) -> Iterator[File]:
    """Render the files."""

    def _renderfiles(paths: Iterable[Path]) -> Iterator[File]:
        for path in paths:
            name = render(path, bindings).name
            if not name:
                continue

            if "/" in name or "\\" in name or name in (".", ".."):
                raise RuntimeError(
                    f"invalid component {name!r} from {path.name!r} in {path}"
                )

            if path.is_dir():
                yield from _renderfiles(path.iterdir())
            else:
                yield render(loadfile(path), bindings)

    return _renderfiles(paths)



"""Mounting filesystems."""
import pathlib
from collections.abc import Callable
from typing import Optional

from cutty.filesystems.domain.filesystem import Filesystem
from cutty.repositories.domain.revisions import Revision


Mounter = Callable[[pathlib.Path, Optional[Revision]], Filesystem]


def unversioned_mounter(filesystem: Callable[[pathlib.Path], Filesystem]) -> Mounter:
    """Return a mounter that raises when a revision is passed."""

    def _mount(storage: pathlib.Path, revision: Optional[Revision]) -> Filesystem:
        if revision is not None:
            raise RuntimeError(f"filesystem does not support revisions, got {revision}")
        return filesystem(storage)

    return _mount



"""Utility for handling an exception by raising another one."""
from typing import NoReturn
from typing import Union

from cutty.util.exceptionhandlers import ExceptionHandler
from cutty.util.exceptionhandlers import exceptionhandler


def reraise(
    exception: BaseException,
    *,
    when: Union[type[BaseException], tuple[type[BaseException], ...]] = ()
) -> ExceptionHandler:
    """Raise the given exception when any of the specified exceptions is caught."""
    if not isinstance(when, tuple):
        when = (when,)
    elif not when:
        when = (Exception,)

    @exceptionhandler(*when)
    def _(other: BaseException) -> NoReturn:
        raise exception

    return _



"""Update a project with changes from its Cookiecutter template."""
import hashlib
import json
import tempfile
from collections.abc import Iterator
from collections.abc import Mapping
from pathlib import Path
from types import MappingProxyType
from typing import Optional

import pygit2

from cutty.compat.contextlib import contextmanager
from cutty.filestorage.adapters.observers.git import commit
from cutty.filestorage.adapters.observers.git import LATEST_BRANCH
from cutty.filestorage.adapters.observers.git import UPDATE_MESSAGE
from cutty.services.create import create


def getprojecttemplate(projectdir: Path) -> str:
    """Return the location of the project template."""
    context = getprojectcontext(projectdir)
    return context["_template"]


def getprojectcontext(projectdir: Path) -> dict[str, str]:
    """Return the Cookiecutter context of the project."""
    text = (projectdir / ".cookiecutter.json").read_text()
    data = json.loads(text)
    return {
        key: value
        for key, value in data.items()
        if isinstance(key, str) and isinstance(value, str)
    }


def checkoutemptytree(repositorypath: Path) -> None:
    """Check out an empty tree from the repository."""
    repository = pygit2.Repository(repositorypath)
    oid = repository.TreeBuilder().write()
    repository.checkout_tree(repository[oid])


@contextmanager
def createworktree(
    repositorypath: Path,
    branch: str,
    *,
    checkout: bool = True,
) -> Iterator[Path]:
    """Create a worktree for the branch in the repository."""
    repository = pygit2.Repository(repositorypath)

    with tempfile.TemporaryDirectory() as directory:
        name = hashlib.blake2b(branch.encode(), digest_size=32).hexdigest()
        path = Path(directory) / name
        worktree = repository.add_worktree(name, path, repository.branches[branch])

        if not checkout:
            # Emulate `--no-checkout` by checking out an empty tree after the fact.
            checkoutemptytree(path)

        yield path

    # Prune with `force=True` because libgit2 thinks `worktree.path` still exists.
    # https://github.com/libgit2/libgit2/issues/5280
    worktree.prune(True)


def cherrypick(repositorypath: Path, reference: str) -> None:
    """Cherry-pick the commit onto the current branch."""
    repository = pygit2.Repository(repositorypath)
    oid = repository.references[reference].resolve().target
    repository.cherrypick(oid)

    if repository.index.conflicts:
        paths = {
            side.path
            for _, ours, theirs in repository.index.conflicts
            for side in (ours, theirs)
            if side is not None
        }
        raise RuntimeError(f"Merge conflicts: {', '.join(paths)}")

    commit(repository, message=UPDATE_MESSAGE)
    repository.state_cleanup()


def update(
    *,
    projectdir: Optional[Path] = None,
    extra_context: Mapping[str, str] = MappingProxyType({}),
    no_input: bool = False,
) -> None:
    """Update a project with changes from its Cookiecutter template."""
    if projectdir is None:
        projectdir = Path.cwd()

    template = getprojecttemplate(projectdir)
    context = getprojectcontext(projectdir)

    with createworktree(projectdir, LATEST_BRANCH, checkout=False) as worktree:
        create(
            template,
            outputdir=worktree,
            outputdirisproject=True,
            extra_context={**context, **extra_context},
            no_input=no_input,
        )

    cherrypick(projectdir, f"refs/heads/{LATEST_BRANCH}")



"""Filesystem-agnostic path."""
from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from typing import overload
from typing import TypeVar
from typing import Union


PurePathT = TypeVar("PurePathT", bound="PurePath")


@dataclass(frozen=True)
class PurePath:
    """Location in a filesystem."""

    parts: tuple[str, ...]

    def __init__(self, *parts: str) -> None:
        """Initialize."""
        object.__setattr__(self, "parts", parts)

    def _withparts(self: PurePathT, *parts: str) -> PurePathT:
        """Create a path with the given parts."""
        return self.__class__(*parts)

    def __str__(self) -> str:
        """Return a readable representation."""
        return "/".join(self.parts)

    def __truediv__(self: PurePathT, part: str) -> PurePathT:
        """Return a path with the part appended."""
        return self.joinpath(part)

    @property
    def name(self) -> str:
        """The final path component, if any."""
        return self.parts[-1] if self.parts else ""

    @property
    def stem(self) -> str:
        """The final path component, minus its last suffix."""
        name = self.name
        index = name.rfind(".")
        return name[:index] if 0 < index < len(name) - 1 else name

    @property
    def parent(self: PurePathT) -> PurePathT:
        """Return the parent of this path."""
        return self._withparts(*self.parts[:-1]) if self.parts else self

    @property
    def parents(self: PurePathT) -> Sequence[PurePathT]:
        """Return the ancestors of the path."""
        return Parents(self)

    def joinpath(self: PurePathT, *parts: str) -> PurePathT:
        """Return a path with the parts appended."""
        return self._withparts(*self.parts, *parts)


class Parents(Sequence[PurePathT]):
    """Sequence-like access to the logical ancestors of a path."""

    def __init__(self, path: PurePathT) -> None:
        """Initialize."""
        self.path = path

    def __len__(self) -> int:
        """Return the number of parents."""
        return len(self.path.parts)

    @overload
    def __getitem__(self, index: int) -> PurePathT:  # noqa: D105
        ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[PurePathT]:  # noqa: D105
        ...

    def __getitem__(
        self, index: Union[int, slice]
    ) -> Union[PurePathT, Sequence[PurePathT]]:
        """Return the nth parent."""
        if isinstance(index, slice):
            indices = index.indices(len(self))
            return tuple(self[index] for index in range(*indices))

        if index >= len(self) or index < -len(self):
            raise IndexError(index)

        parts = self.path.parts[: -(index + 1)]
        return self.path._withparts(*parts)



"""Exception handlers as first-class citizens.

The utilities in this module allow the creation of exception handlers as
first-class objects that can be passed around and combined.

Consider the following imaginary application:

>>> class ApplicationError(Exception):
...     pass
...
>>> def main():
...     pass

Conventionally, application errors would be handled using a ``try...except``
block:

>>> try:
...     main()
... except ApplicationError as error:
...     sys.exit(f"fatal: {error}")

Instead, this module lets you create an exception handler object:

>>> @exceptionhandler(ApplicationError)
... def fatal(error):
...     sys.exit(f"fatal: {error}")

The :func:`exceptionhandler` decorator creates an instance of
:class:`ExceptionHandler`, a context manager that invokes the wrapped function
when an exception of type ApplicationError is caught.

The exception handler can now be used anywhere in the code, like this:

>>> with fatal:
...     main()

Exception handlers can also be used as decorators to apply exception handling to
a function. So instead of wrapping every call to ``main`` in a ``with``
statement, you could apply the exception handler to ``main`` once:

>>> @fatal
... def main():
...     ...

If your code uses type annotations, you can omit the arguments to
``exceptionhandler``, and define the handler like this:

>>> @exceptionhandler
... def fatal(error: ApplicationError):
...     sys.exit(f"fatal: {error}")

An important difference between an ``except`` clause and an ``ExceptionHandler``
is that ``except`` clauses suppress exceptions by default, and require an
explicit ``raise`` statement to re-raise the exception. By contrast, exception
handlers re-raise exceptions by default, and require an explicit ``return True``
statement to suppress the exception:

>>> @exceptionhandler(Exception)
... def suppress(exception):
...     return True

Exception handlers can be combined using the ``<<`` and ``>>`` operators:

>>> with fatal >> suppress:
...     ...

The operator points in the direction of exception flow, so in this example
``fatal`` sees exceptions before ``suppress``; in other words, the code exits on
application errors, while any remaining exceptions are silently discarded.

In general, ``a >> b`` is equivalent to ``b << a``. For example, the code above
could also have been written like this:

>>> with suppress << fatal:
...     ...

This style is closer to how the code would look with multiple context managers:

>>> with suppress, fatal:
...     ...

While this form is shorter, composing exception handlers explicitly gives us an
``ExceptionHandler`` object. The combined exception handler can be stored in a
variable, passed around, or combined with further handlers:

>>> handler = fatal >> suppress

This module defines only one actual exception handler of its own:
``nullhandler`` is a noop that does not handle any exceptions. A practical use
of ``nullhandler`` is when exception handling should be optional, letting you
either pass the real handler or ``nullhandler``.

In algebraic terms, ``nullhandler`` is the identity element of composition
(``<<`` and ``>>``): Combining any handler with ``nullhandler`` gives you a
handler that behaves exactly like the original handler. As composition is also
associative, this makes exception handlers with composition a simple monoid_.

.. _monoid: https://en.wikipedia.org/wiki/Monoid
"""
from __future__ import annotations

import contextlib
from collections.abc import Callable
from types import TracebackType
from typing import Generic
from typing import get_type_hints
from typing import Optional
from typing import overload
from typing import TypeVar
from typing import Union


__all__ = [
    "ExceptionHandler",
    "exceptionhandler",
    "nullhandler",
]


class ExceptionHandler(contextlib.ContextDecorator):
    """Context manager for handling exceptions.

    Apply exception handlers to a block of code using a with block::

        with myhandler:
            ...

    Note that there are no parentheses, because there is no need to invoke the
    exception handler. Exception handlers are `reentrant context managers`_,
    which means that you do not need to create them afresh for each use, even
    when they are used multiple times in nested ``with`` statements.

    .. _reentrant context managers:
       https://docs.python.org/3/library/contextlib.html
       #single-use-reusable-and-reentrant-context-managers

    Exception handlers can also be used as decorators::

        @myhandler
        def f():
            ...

    This is equivalent to enclosing the body of the function in a ``with``
    block::

        def f():
            with myhandler:
                ...

    This class should not be instantiated directly. Use the
    :func:`exceptionhandler` decorator to create an exception handler.
    """

    def __enter__(self) -> None:
        """Enter the runtime context."""

    def __exit__(
        self,
        exception_type: Optional[type[BaseException]],
        exception: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        """Exit the runtime context."""
        return None

    def __lshift__(self, other: ExceptionHandler) -> ExceptionHandler:
        """Compose this exception handler with another one.

        The expression ``a << b`` is equivalent to these nested ``with``
        blocks::

            with a:
                with b:
                    ...

        As a mnemonic, the operator points in the direction of exception flow.
        In ``a << b``, the right operand sees the exception first.
        """
        return _Compose(self, other)

    def __rshift__(self, other: ExceptionHandler) -> ExceptionHandler:
        """Compose this exception handler with another one.

        The expression ``a >> b`` is equivalent to these nested ``with``
        blocks::

            with b:
                with a:
                    ...

        As a mnemonic, the operator points in the direction of exception flow.
        In ``a >> b``, the left operand sees the exception first.
        """
        return other << self


nullhandler = ExceptionHandler()


class _Compose(ExceptionHandler):
    """Exception handler composed of other exception handlers."""

    def __init__(self, first: ExceptionHandler, second: ExceptionHandler) -> None:
        """Initialize."""
        # Reduce object count in expressions like ``a << b << c``.
        self.stacks: list[contextlib.ExitStack] = []
        self.handlers: list[ExceptionHandler] = [
            handler
            for arg in (first, second)
            for handler in (arg.handlers if isinstance(arg, _Compose) else [arg])
        ]

    def __enter__(self) -> None:
        """Enter the runtime context."""
        stack = contextlib.ExitStack()
        self.stacks.append(stack)

        for handler in self.handlers:
            stack.enter_context(handler)

        stack.__enter__()

    def __exit__(
        self,
        exception_type: Optional[type[BaseException]],
        exception: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        """Exit the runtime context."""
        stack = self.stacks.pop()
        return stack.__exit__(exception_type, exception, traceback)


E = TypeVar("E", bound=BaseException, contravariant=True)


_Callback = Callable[[E], Optional[bool]]


class _Decorator(Generic[E], ExceptionHandler):
    """Exception handler wrapping a callback function.

    This class is a helper for the @exceptionhandler decorator.
    """

    def __init__(
        self,
        callback: _Callback[E],
        *exception_types: type[BaseException],
    ) -> None:
        """Initialize."""
        self.callback = callback
        self.exception_types = exception_types

    def __exit__(
        self,
        exception_type: Optional[type[BaseException]],
        exception: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        """Exit the context."""
        return (
            self.callback(exception)  # type: ignore[arg-type]
            if any(
                isinstance(exception, exception_type)
                for exception_type in self.exception_types
            )
            else None
        )


def _exceptionhandler(
    *exception_types: type[BaseException],
) -> Callable[[_Callback[E]], ExceptionHandler]:
    """Decorator factory implementing @exceptionhandler."""

    def _decorator(callback: _Callback[E]) -> ExceptionHandler:
        if exception_types:
            return _Decorator(callback, *exception_types)

        try:
            exception_type = next(
                hint
                for key, hint in get_type_hints(callback).items()
                if key != "return"
            )
        except StopIteration:
            raise TypeError(f"missing type annotation on {callback}") from None

        return _Decorator(callback, exception_type)

    return _decorator


@overload
def exceptionhandler(__callback: _Callback[E]) -> ExceptionHandler:
    ...


@overload
def exceptionhandler(
    *exception_types: type[BaseException],
) -> Callable[[_Callback[E]], ExceptionHandler]:
    ...


def exceptionhandler(
    *args: Union[_Callback[E], type[BaseException]],
) -> Union[ExceptionHandler, Callable[[_Callback[E]], ExceptionHandler]]:
    """Decorator for creating exception handlers.

    This decorator creates an exception handler that invokes the wrapped
    function when an exception of the specified type is caught::

        @exceptionhandler(SomeError)
        def handler(error):
            return True

    Multiple exception types may be specified, or none. In the latter case, the
    exception type is derived from the type annotation for the function
    parameter. Parentheses for the decorator are optional in this case. The
    following example is equivalent to the first one::

        @exceptionhandler
        def handler(error: SomeError) -> bool:
            return True

    A return value of True suppresses the exception. If the function returns
    False or None, the exception is reraised when the handler returns::

        @exceptionhandler
        def print_handler(exception: SomeException) -> None:
            print(exception)  # the exception is printed but not swallowed

    Another option is to raise another exception::

        @exceptionhandler
        def exit_handler(exception: SomeException) -> NoReturn:
            raise SystemExit(str(exception))

    """
    if all(isinstance(arg, type) and issubclass(arg, BaseException) for arg in args):
        exception_types: tuple[type[BaseException]] = args  # type: ignore[assignment]
        return _exceptionhandler(*exception_types)

    callback: _Callback[E]
    [callback] = args  # type: ignore[assignment]
    return _exceptionhandler()(callback)



"""Rendering with Jinja."""
from __future__ import annotations

import contextlib
from collections.abc import Callable
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Any
from typing import Optional
from typing import Union

import jinja2

from cutty.filesystems.domain.path import Path
from cutty.templates.domain.bindings import Binding
from cutty.templates.domain.render import GenericRenderer
from cutty.util.reraise import reraise


@dataclass
class TemplateExtensionNotFoundError(Exception):
    """The template extension was not found."""

    extension: str


@dataclass
class TemplateExtensionTypeError(Exception):
    """The template extension does not have the expected type."""

    extension: str
    type: str


def import_object(import_path: str) -> Any:
    """Import the object at the given import path.

    Import paths consist of the dotted module name, optionally followed by a
    colon or dot, and the module attribute at which the object is located.

    For example:

    - ``json``
    - ``os.path``
    - ``xml.sax.saxutils:escape``
    - ``xml.sax.saxutils.escape``

    This function mirrors the implementation of ``jinja2.utils.import_string``.
    """
    if ":" in import_path:
        module_name, _, attribute = import_path.rpartition(":")
    elif "." in import_path:
        module_name, _, attribute = import_path.rpartition(".")
    else:
        return __import__(import_path)

    module = __import__(module_name, None, None, [attribute])
    return getattr(module, attribute)


def load_extension(import_path: str) -> type[jinja2.ext.Extension]:
    """Import a Jinja extension from the specified path."""
    with reraise(TemplateExtensionNotFoundError(import_path)):
        extension = import_object(import_path)

    if not (
        isinstance(extension, type) and issubclass(extension, jinja2.ext.Extension)
    ):
        raise TemplateExtensionTypeError(import_path, str(type(extension)))

    return extension


def splitpath(pathstr: str) -> tuple[str, ...]:
    """Split a path into segments and perform a sanity check.

    If it detects '..' in the path it will raise a `TemplateNotFound` error.

    source: jinja2.loaders.split_template_path
    """
    # TODO: Add string parsing to PurePath?
    # TODO: Add common validation function?
    # (see cutty.templates.domain.render.renderfiles)
    import os

    parts = tuple(part for part in pathstr.split("/") if part and part != os.curdir)

    if any(
        any(sep in part for sep in (os.sep, os.altsep) if sep) or part == os.pardir
        for part in parts
    ):
        raise jinja2.TemplateNotFound(pathstr)

    return parts


class _FilesystemLoader(jinja2.BaseLoader):
    """Load templates from a directory in the file system."""

    def __init__(self, *, searchpath: Iterable[Path]):
        """Initialize."""
        self.searchpath = tuple(searchpath)

    def get_source(
        self, environment: jinja2.Environment, template: str
    ) -> tuple[str, str, Callable[[], bool]]:
        """Get the template source, filename and reload helper for a template."""
        parts = splitpath(template)
        for searchpath in self.searchpath:
            path = searchpath.joinpath(*parts)
            with contextlib.suppress(FileNotFoundError):
                return path.read_text(), str(path), lambda: True

        raise jinja2.TemplateNotFound(template)


def createjinjarenderer(
    *,
    searchpath: Iterable[Path],
    context_prefix: Optional[str] = None,
    extra_context: Optional[dict[str, Any]] = None,
    extensions: Iterable[Union[str, type[jinja2.ext.Extension]]] = (),
) -> GenericRenderer[str]:
    """Create a renderer using Jinja."""
    extensions = [
        load_extension(extension) if isinstance(extension, str) else extension
        for extension in extensions
    ]

    environment = jinja2.Environment(  # noqa: S701
        loader=_FilesystemLoader(searchpath=searchpath),
        extensions=extensions,
        keep_trailing_newline=True,
        undefined=jinja2.StrictUndefined,
    )

    def rendertext(text: str, bindings: Iterable[Binding]) -> str:
        """Render the text using Jinja."""
        template = environment.from_string(text)
        context = {binding.name: binding.value for binding in bindings}

        if extra_context is not None:
            context.update(extra_context)

        if context_prefix is not None:
            context = {context_prefix: context}

        return template.render(context)

    return rendertext



"""Disk-based file storage."""
import contextlib
import enum
import pathlib
from collections.abc import Callable
from typing import Optional

from cutty.filestorage.domain.files import Executable
from cutty.filestorage.domain.files import File
from cutty.filestorage.domain.files import RegularFile
from cutty.filestorage.domain.files import SymbolicLink
from cutty.filestorage.domain.storage import FileStorage
from cutty.filesystems.domain.purepath import PurePath


class FileExistsPolicy(enum.Enum):
    """What to do when a file already exists."""

    RAISE = enum.auto()
    OVERWRITE = enum.auto()
    SKIP = enum.auto()

    def check(self, path: pathlib.Path) -> Optional[bool]:
        """Check the policy for a given path.

        Args:
            path: The path to be checked against the policy.

        Returns:
            ``True`` to overwrite the path, ``False`` to skip the path.

        Raises:
            FileExistsError: The policy does not permit overwriting the path.
        """
        if self is FileExistsPolicy.RAISE:
            raise FileExistsError(f"{path} already exists")

        return self is FileExistsPolicy.OVERWRITE


class DiskFileStorage(FileStorage):
    """Disk-based file storage."""

    def __init__(
        self,
        root: pathlib.Path,
        *,
        fileexists: FileExistsPolicy = FileExistsPolicy.RAISE,
    ) -> None:
        """Initialize."""
        super().__init__()
        self.root = root
        self.fileexists = fileexists
        self.undo: list[Callable[[], None]] = []

    def begin(self) -> None:
        """Begin a storage transaction."""

    def add(self, file: File) -> None:
        """Add the file to the storage."""
        path = self.resolve(file.path)
        if not path.exists():
            self._storefile(file, path, overwrite=False)
        elif self.fileexists.check(path):
            self._storefile(file, path, overwrite=True)

    def resolve(self, path: PurePath) -> pathlib.Path:
        """Return the filesystem location."""
        return self.root.joinpath(*path.parts)

    def _storefile(
        self,
        file: File,
        path: pathlib.Path,
        *,
        overwrite: bool = False,
    ) -> None:
        """Store the file at the given path on disk."""
        # OVERWRITING.
        #
        # These operations are allowed:
        #
        # - Overwrite a regular file with another regular file.
        # - Overwrite a symbolic link with a regular file.
        # - Overwrite a symbolic link with another symbolic link.
        #
        # These operations raise exceptions:
        #
        # - Do not overwrite a directory with a regular file.
        # - Do not overwrite a directory with a symbolic link.
        # - Do not overwrite a regular file with a symbolic link.
        # - Do not overwrite a regular file with a directory.
        # - Do not overwrite a symbolic link with a directory if the link target is
        #   not a directory. (Symbolic links in the path are followed if they point
        #   to directories.)

        if overwrite and path.is_symlink():
            path.unlink()

        for parent in reversed(path.parents):
            if not parent.is_dir():
                parent.mkdir()
                self.undo.append(parent.rmdir)

        if isinstance(file, RegularFile):
            path.write_bytes(file.blob)

            if not overwrite:
                self.undo.append(path.unlink)

            if isinstance(file, Executable):
                path.chmod(path.stat().st_mode | 0o111)

        elif isinstance(file, SymbolicLink):
            target = pathlib.Path(*file.target.parts)
            path.symlink_to(target)

            if not overwrite:
                self.undo.append(path.unlink)

        else:
            raise TypeError(f"cannot store file of type {type(file)}")

    def commit(self) -> None:
        """Commit all stores."""

    def rollback(self) -> None:
        """Rollback all stores."""
        for action in reversed(self.undo):
            with contextlib.suppress(Exception):
                action()



"""Git-based filesystem using libgit2."""
import pathlib
from collections.abc import Iterator

import pygit2

from cutty.filesystems.domain.filesystem import Access
from cutty.filesystems.domain.nodefs import FilesystemNode
from cutty.filesystems.domain.nodefs import NodeFilesystem
from cutty.filesystems.domain.purepath import PurePath


class GitFilesystemNode(FilesystemNode):
    """A node in a git filesystem."""

    def __init__(self, node: pygit2.Object) -> None:
        """Initialize."""
        self.node = node

    def is_dir(self) -> bool:
        """Return True if the node is a directory."""
        return isinstance(self.node, pygit2.Tree)

    def is_file(self) -> bool:
        """Return True if the node is a regular file."""
        return (
            isinstance(self.node, pygit2.Blob)
            and self.node.filemode != pygit2.GIT_FILEMODE_LINK
        )

    def is_symlink(self) -> bool:
        """Return True if the node is a symbolic link."""
        return (
            isinstance(self.node, pygit2.Blob)
            and self.node.filemode == pygit2.GIT_FILEMODE_LINK
        )

    def read_bytes(self) -> bytes:
        """Return the file contents."""
        data: bytes = self.node.data
        return data

    def read_text(self) -> str:
        """Return the file contents."""
        return self.read_bytes().decode()

    def readlink(self) -> PurePath:
        """Return the link target."""
        target: str = self.node.data.decode(errors="surrogateescape")
        parts = pathlib.PurePosixPath(target).parts
        return PurePath(*parts)

    def iterdir(self) -> Iterator[str]:
        """Iterate over the directory entries."""
        for entry in self.node:
            yield entry.name

    def __truediv__(self, entry: str) -> FilesystemNode:
        """Return the given directory entry."""
        try:
            return GitFilesystemNode(self.node / entry)
        except KeyError:
            raise FileNotFoundError()

    def access(self, mode: Access) -> bool:
        """Return True if the user can access the node."""
        return (
            Access.EXECUTE not in mode
            or self.node.filemode == pygit2.GIT_FILEMODE_BLOB_EXECUTABLE
        )


class GitFilesystem(NodeFilesystem):
    """Git-based filesystem."""

    def __init__(self, repository: pathlib.Path, ref: str = "HEAD") -> None:
        """Inititalize."""
        repo = pygit2.Repository(repository)
        tree = repo.revparse_single(ref).peel(pygit2.Tree)
        self.root = GitFilesystemNode(tree)



"""Providers for git repositories."""
import pathlib
from typing import Optional

import pygit2

from cutty.filesystems.adapters.git import GitFilesystem
from cutty.repositories.adapters.fetchers.git import gitfetcher
from cutty.repositories.domain.providers import localprovider
from cutty.repositories.domain.providers import remoteproviderfactory
from cutty.repositories.domain.revisions import Revision


def match(path: pathlib.Path) -> bool:
    """Return True if the path is a git repository."""
    repository = pygit2.discover_repository(path)
    if repository is None:
        return False

    repositorypath = pathlib.Path(repository)
    return path in (repositorypath, repositorypath.parent)


def mount(path: pathlib.Path, revision: Optional[Revision]) -> GitFilesystem:
    """Return a filesystem tree for the given revision.

    This function returns the root of a Git filesystem for the given
    revision. If ``revision`` is None, HEAD is used instead.
    """
    return GitFilesystem(path, *filter(None, [revision]))


localgitprovider = localprovider(match=match, mount=mount)
gitproviderfactory = remoteproviderfactory(fetch=[gitfetcher], mount=mount)



"""Filesystem implementation for ZIP archives using zipfile."""
import pathlib
import stat
import zipfile
from collections.abc import Iterator

from cutty.filesystems.domain.filesystem import Access
from cutty.filesystems.domain.nodefs import FilesystemNode
from cutty.filesystems.domain.nodefs import NodeFilesystem
from cutty.filesystems.domain.purepath import PurePath


def _fromaccess(access: Access) -> int:
    mapping = {
        Access.READ: stat.S_IRUSR,
        Access.WRITE: stat.S_IWUSR,
        Access.EXECUTE: stat.S_IXUSR,
    }
    return sum([mapping[flag] for flag in Access if flag and flag in access])


def _getfilemode(zippath: zipfile.Path) -> int:
    info: zipfile.ZipInfo
    info = zippath.root.getinfo(zippath.at)  # type: ignore[attr-defined]
    return info.external_attr >> 16


class ZipFilesystemNode(FilesystemNode):
    """A node in a ZIP filesystem."""

    def __init__(self, node: zipfile.Path) -> None:
        """Initialize."""
        self.node = node

    def is_dir(self) -> bool:
        """Return True if the node is a directory."""
        return self.node.is_dir()

    def is_file(self) -> bool:
        """Return True if the node is a regular file."""
        return self.node.is_file()

    def is_symlink(self) -> bool:
        """Return True if the node is a symbolic link."""
        return False

    def read_bytes(self) -> bytes:
        """Return the file contents."""
        return self.node.read_bytes()

    def read_text(self) -> str:
        """Return the file contents."""
        return self.node.read_text()

    def readlink(self) -> PurePath:
        """Return the link target."""
        raise NotImplementedError()  # pragma: no cover

    def iterdir(self) -> Iterator[str]:
        """Iterate over the directory entries."""
        for entry in self.node.iterdir():
            yield entry.name

    def __truediv__(self, entry: str) -> FilesystemNode:
        """Return the given directory entry."""
        node = self.node / entry

        if not node.exists():
            raise FileNotFoundError()

        return ZipFilesystemNode(node)

    def access(self, mode: Access) -> bool:
        """Return True if the user can access the node."""
        return not mode or bool(_getfilemode(self.node) & _fromaccess(mode))


class ZipFilesystem(NodeFilesystem):
    """ZIP filesystem."""

    def __init__(self, path: pathlib.Path) -> None:
        """Inititalize."""
        node = zipfile.Path(path)
        self.root = ZipFilesystemNode(node)



"""Bindings."""
from dataclasses import dataclass
from typing import Any
from typing import Generic
from typing import TypeVar


T_co = TypeVar("T_co", covariant=True)


@dataclass(frozen=True)
class GenericBinding(Generic[T_co]):
    """A binding associates a name with a value."""

    name: str
    value: T_co


Binding = GenericBinding[Any]



"""Console prompts implemented using the click package."""
import json
from typing import Any
from typing import Optional

import click

from cutty.templates.domain.binders import bind
from cutty.templates.domain.binders import binddefault
from cutty.templates.domain.bindings import Binding
from cutty.templates.domain.variables import Variable


def textprompt(variable: Variable) -> Binding:
    """Prompt for name and return the entered value or given default."""
    # typeshed incorrectly requires Optional[str] for `default`
    value = click.prompt(
        variable.name,
        default=variable.default,
    )
    return bind(variable, value)


def choiceprompt(variable: Variable) -> Binding:
    """Prompt to choose from several values for the given name."""
    if not variable.choices:
        raise ValueError("variable with empty choices")

    choices = {str(number): value for number, value in enumerate(variable.choices, 1)}

    lines = [
        f"Select {variable.name}:",
        *[f"{number} - {value}" for number, value in choices.items()],
        "Choose from {}".format(", ".join(choices.keys())),
    ]

    choice = click.prompt(
        "\n".join(lines),
        type=click.Choice(list(choices)),
        default="1",
        show_choices=False,
    )

    return bind(variable, choices[choice])


def jsonprompt(variable: Variable) -> Binding:
    """Prompt to provide a dictionary of data."""

    def _load_json_dict(value: Optional[str]) -> dict[str, Any]:
        """Load entered value as a JSON dict."""
        if value == "default":
            result: dict[str, Any] = variable.default
            return result

        assert value is not None  # noqa: S101

        try:
            result = json.loads(value)
        except Exception as error:
            raise click.UsageError(f"unable to decode to JSON: {error}")

        if not isinstance(result, dict):
            raise click.UsageError("requires JSON object")

        return result

    value = click.prompt(
        variable.name,
        default="default",
        type=click.STRING,
        value_proc=_load_json_dict,
    )

    return bind(variable, value)


def prompt(variable: Variable) -> Binding:
    """Bind a variable by prompting."""
    if not variable.interactive:
        return binddefault(variable)

    if variable.choices:
        return choiceprompt(variable)

    if issubclass(variable.type, dict):
        return jsonprompt(variable)

    return textprompt(variable)



"""File storage abstraction."""
import abc
from types import TracebackType
from typing import Optional
from typing import TypeVar

from cutty.filestorage.domain.files import File


T = TypeVar("T", bound="FileStorage")


class FileStorage(abc.ABC):
    """Interface for file storage implementations."""

    @abc.abstractmethod
    def begin(self) -> None:
        """Begin a storage transaction."""

    @abc.abstractmethod
    def add(self, file: File) -> None:
        """Add the file to the storage."""

    @abc.abstractmethod
    def commit(self) -> None:
        """Commit all stores."""

    @abc.abstractmethod
    def rollback(self) -> None:
        """Rollback all stores."""

    def __enter__(self: T) -> T:
        """Enter the runtime context."""
        self.begin()
        return self

    def __exit__(
        self,
        exception_type: Optional[type[BaseException]],
        exception: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Exit the runtime context."""
        if exception is None:
            self.commit()
        else:
            self.rollback()



"""Default filesystem implementation based on nodes."""
from __future__ import annotations

import abc
import functools
from collections.abc import Iterator

from cutty.filesystems.domain.filesystem import Access
from cutty.filesystems.domain.filesystem import Filesystem
from cutty.filesystems.domain.purepath import PurePath


class FilesystemNode(abc.ABC):
    """A node in the filesystem."""

    @abc.abstractmethod
    def is_dir(self) -> bool:
        """Return True if the node is a directory."""

    @abc.abstractmethod
    def is_file(self) -> bool:
        """Return True if the node is a regular file."""

    @abc.abstractmethod
    def is_symlink(self) -> bool:
        """Return True if the node is a symbolic link."""

    @abc.abstractmethod
    def read_bytes(self) -> bytes:
        """Return the file contents."""

    @abc.abstractmethod
    def read_text(self) -> str:
        """Return the file contents."""

    @abc.abstractmethod
    def readlink(self) -> PurePath:
        """Return the link target."""

    @abc.abstractmethod
    def iterdir(self) -> Iterator[str]:
        """Iterate over the directory entries."""

    @abc.abstractmethod
    def __truediv__(self, entry: str) -> FilesystemNode:
        """Return the given directory entry."""

    @abc.abstractmethod
    def access(self, mode: Access) -> bool:
        """Return True if the user can access the node."""


class InvalidArgumentError(Exception):
    """The filesystem operation received an invalid argument."""


class NodeFilesystem(Filesystem):
    """A partial filesystem implementation based on nodes."""

    root: FilesystemNode

    @functools.cache
    def lookup(self, path: PurePath) -> FilesystemNode:
        """Return the filesystem node located at the given path."""
        originalpath = path
        nodes = [self.root]

        def _lookup(path: PurePath) -> FilesystemNode:
            """Return the filesystem node for the given path."""
            for part in path.parts:
                if not nodes[-1].is_dir():
                    raise NotADirectoryError(f"not a directory: {path}")

                if part == ".":
                    continue

                if part == "..":
                    if len(nodes) > 1:
                        nodes.pop()
                    continue  # pragma: no cover

                node = self._lookup_entry(nodes[-1], part, path=originalpath)

                if node.is_symlink():
                    target = node.readlink()
                    node = _lookup(target)

                nodes.append(node)

            return nodes[-1]

        return _lookup(path)

    @classmethod
    def _lookup_entry(
        cls, node: FilesystemNode, entry: str, *, path: PurePath
    ) -> FilesystemNode:
        """Lookup a directory entry."""
        try:
            return node / entry
        except FileNotFoundError:
            raise FileNotFoundError(f"file not found: {path}")

    def _lookup_symlink(self, path: PurePath) -> FilesystemNode:
        """Lookup the path without resolving the final path component."""
        parent = self.lookup(path.parent)

        if not parent.is_dir():
            raise NotADirectoryError(f"not a directory: {path}")

        return self._lookup_entry(parent, path.name, path=path)

    def is_dir(self, path: PurePath) -> bool:
        """Return True if this is a directory."""
        try:
            node = self.lookup(path)
        except (FileNotFoundError, NotADirectoryError):
            return False
        else:
            return node.is_dir()

    def iterdir(self, path: PurePath) -> Iterator[str]:
        """Iterate over the files in this directory."""
        node = self.lookup(path)

        if not node.is_dir():
            raise NotADirectoryError(f"not a directory: {path}")

        yield from node.iterdir()

    def is_file(self, path: PurePath) -> bool:
        """Return True if this is a regular file (or a symlink to one)."""
        try:
            node = self.lookup(path)
        except (FileNotFoundError, NotADirectoryError):
            return False
        else:
            return node.is_file()

    def read_bytes(self, path: PurePath) -> bytes:
        """Return the contents of this file."""
        node = self.lookup(path)

        if node.is_dir():
            raise IsADirectoryError(f"is a directory: {path}")

        return node.read_bytes()

    def read_text(self, path: PurePath) -> str:
        """Return the contents of this file."""
        node = self.lookup(path)

        if node.is_dir():
            raise IsADirectoryError(f"is a directory: {path}")

        return node.read_text()

    def is_symlink(self, path: PurePath) -> bool:
        """Return True if this is a symbolic link."""
        try:
            node = self._lookup_symlink(path)
        except (FileNotFoundError, NotADirectoryError):
            return False
        else:
            return node.is_symlink()

    def readlink(self, path: PurePath) -> PurePath:
        """Return the target of a symbolic link."""
        node = self._lookup_symlink(path)

        if node.is_dir():
            raise IsADirectoryError(f"is a directory: {path}")

        if not node.is_symlink():
            raise InvalidArgumentError(f"not a symbolic file: {path}")

        return node.readlink()

    def access(self, path: PurePath, mode: Access) -> bool:
        """Return True if the user can access the path."""
        try:
            node = self.lookup(path)
        except (FileNotFoundError, NotADirectoryError):
            return False
        else:
            return node.access(mode)



"""Provider registry."""
import shutil

from cutty.repositories.adapters.providers.disk import diskprovider
from cutty.repositories.adapters.providers.git import gitproviderfactory
from cutty.repositories.adapters.providers.git import localgitprovider
from cutty.repositories.adapters.providers.mercurial import hgproviderfactory
from cutty.repositories.adapters.providers.zip import localzipprovider
from cutty.repositories.adapters.providers.zip import zipproviderfactory
from cutty.repositories.domain.providers import constproviderfactory as factory


defaultproviderregistry = {
    "localzip": factory(localzipprovider),
    "localgit": factory(localgitprovider),
    "local": factory(diskprovider),
    "zip": zipproviderfactory,
    "git": gitproviderfactory,
}

if shutil.which("hg") is not None:  # pragma: no cover
    defaultproviderregistry["hg"] = hgproviderfactory



"""Rendering Cookiecutter templates."""
import fnmatch
from collections.abc import Sequence
from typing import Any

from binaryornot.helpers import is_binary_string

from cutty.filestorage.domain.files import Executable
from cutty.filestorage.domain.files import RegularFile
from cutty.filesystems.domain.path import Path
from cutty.templates.adapters.cookiecutter.extensions import DEFAULT_EXTENSIONS
from cutty.templates.adapters.jinja import createjinjarenderer
from cutty.templates.domain.bindings import Binding
from cutty.templates.domain.config import Config
from cutty.templates.domain.render import asrendercontinuation
from cutty.templates.domain.render import createrenderer
from cutty.templates.domain.render import defaultrenderregistry
from cutty.templates.domain.render import Renderer
from cutty.templates.domain.render import RenderRegistry
from cutty.templates.domain.render import T


def is_binary(blob: bytes) -> bool:
    """Return True if the blob contains binary data."""
    result: bool = is_binary_string(blob[:1024])
    return result


def asstringlist(settings: dict[str, Any], name: str) -> list[str]:
    """Return a setting as a list of strings."""
    value = settings.get(name, [])
    assert isinstance(value, list) and all(  # noqa: S101
        isinstance(item, str) for item in value
    )
    return value


def renderlist(
    values: list[T], bindings: Sequence[Binding], render: Renderer
) -> list[T]:
    """Render a list."""
    return [render(value, bindings) for value in values]


def renderdict(
    mapping: dict[str, T], bindings: Sequence[Binding], render: Renderer
) -> dict[str, T]:
    """Render a dictionary."""
    return {
        render(key, bindings): render(value, bindings) for key, value in mapping.items()
    }


def registerrenderers(path: Path, config: Config) -> RenderRegistry:  # noqa: C901
    """Register render functions."""
    copy_without_render = asstringlist(config.settings, "_copy_without_render")
    extensions = DEFAULT_EXTENSIONS[:]
    extensions.extend(asstringlist(config.settings, "_extensions"))

    def renderregularfile(
        file: RegularFile, bindings: Sequence[Binding], render: Renderer
    ) -> RegularFile:
        """Render a file."""
        cls = Executable if isinstance(file, Executable) else RegularFile
        path = render(file.path, bindings)

        ancestors = [file.path, *file.path.parents[:-1]]
        for ancestor in reversed(ancestors):
            for pattern in copy_without_render:
                if fnmatch.fnmatch(str(ancestor), pattern):
                    return cls(path, file.blob)

        if is_binary(file.blob):
            return cls(path, file.blob)

        text = file.blob.decode()
        text = render(text, bindings)
        return cls(path, text.encode())

    rendertext = createjinjarenderer(
        searchpath=[path],
        context_prefix="cookiecutter",
        extra_context=config.settings,
        extensions=extensions,
    )

    return {
        str: asrendercontinuation(rendertext),
        list: renderlist,
        dict: renderdict,
        RegularFile: renderregularfile,
    }


def createcookiecutterrenderer(path: Path, config: Config) -> Renderer:
    """Create Cookiecutter renderer."""
    renderregistry = registerrenderers(path, config)
    return createrenderer({**defaultrenderregistry, **renderregistry})



"""Filesystem abstraction.

Usage:

>>> from cutty.filesystems.adapters.dict import DictFilesystem
>>> from cutty.filesystems.domain.path import Path
>>> filesystem = DictFilesystem({"README": "hello"})
>>> path = Path(filesystem=filesystem)
>>> assert path.is_dir()
>>> assert (path / "README").read_text() == "hello"
"""



"""Fetch a repository from the filesystem."""
import pathlib
import shutil
from typing import Optional

from yarl import URL

from cutty.repositories.domain.fetchers import fetcher
from cutty.repositories.domain.locations import aspath
from cutty.repositories.domain.matchers import scheme
from cutty.repositories.domain.revisions import Revision


@fetcher(match=scheme("file"))
def filefetcher(
    url: URL, destination: pathlib.Path, revision: Optional[Revision]
) -> None:
    """Copy a file or directory."""
    if destination.is_dir():
        shutil.rmtree(destination)
    elif destination.exists():
        destination.unlink()

    source = aspath(url)

    if source.is_dir():
        shutil.copytree(source, destination, symlinks=True)
    else:
        shutil.copy2(source, destination, follow_symlinks=False)



"""Files."""
from __future__ import annotations

import dataclasses
from dataclasses import dataclass
from typing import TypeVar

from cutty.filesystems.domain.filesystem import Access
from cutty.filesystems.domain.path import Path
from cutty.filesystems.domain.purepath import PurePath


_FileT = TypeVar("_FileT", bound="File")


@dataclass(frozen=True)
class File:
    """A file."""

    path: PurePath

    def withpath(self: _FileT, path: PurePath) -> _FileT:
        """Return a copy of this file with the specified path."""
        return dataclasses.replace(self, path=path)


@dataclass(frozen=True)
class RegularFile(File):
    """A regular file."""

    blob: bytes


@dataclass(frozen=True)
class Executable(RegularFile):
    """An executable file."""


@dataclass(frozen=True)
class SymbolicLink(File):
    """A symbolic link."""

    target: PurePath


def loadfile(
    path: Path,
    *,
    follow_symlinks: bool = True,
) -> File:
    """Load file from path."""
    if not follow_symlinks and path.is_symlink():
        target = path.readlink()
        return SymbolicLink(path, target)

    if path.is_file():
        cls = Executable if path.access(Access.EXECUTE) else RegularFile
        return cls(path, path.read_bytes())

    message = (
        "broken symlink"
        if path.is_symlink()
        else "directory"
        if path.is_dir()
        else "special file"
        if path.access()
        else "no such file"
    )
    raise RuntimeError(f"{path}: {message}")



"""Command-line interface for updating projects from Cookiecutter templates."""
import pathlib
from typing import Optional

import click

from cutty.entrypoints.cli._main import main as _main
from cutty.entrypoints.cli.create import extra_context_callback
from cutty.services.update import update as service_update


@_main.command()
@click.argument("extra-context", nargs=-1, callback=extra_context_callback)
@click.option(
    "--no-input",
    is_flag=True,
    default=False,
    help="Do not prompt for template variables.",
)
@click.option(
    "-C",
    "--cwd",
    metavar="DIR",
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, path_type=pathlib.Path
    ),
    help="Directory of the generated project.",
)
def update(
    extra_context: dict[str, str],
    no_input: bool,
    cwd: Optional[pathlib.Path],
) -> None:
    """Update a project with changes from its template."""
    service_update(extra_context=extra_context, no_input=no_input, projectdir=cwd)



"""Repository provider for a local directory."""
from cutty.filesystems.adapters.disk import DiskFilesystem
from cutty.repositories.domain.mounters import unversioned_mounter
from cutty.repositories.domain.providers import localprovider


diskprovider = localprovider(
    match=lambda path: path.is_dir(),
    mount=unversioned_mounter(DiskFilesystem),
)



"""Template variables."""
from dataclasses import dataclass
from typing import Any
from typing import Generic
from typing import TypeVar


T_co = TypeVar("T_co", covariant=True)


@dataclass(frozen=True)
class GenericVariable(Generic[T_co]):
    """Template variables are specifications for bindings."""

    name: str
    description: str
    type: type[T_co]
    default: T_co
    choices: tuple[T_co, ...]
    interactive: bool


Variable = GenericVariable[Any]


def validate(value: Any, variable: Variable) -> None:
    """Raise ValueError if the value is not compatible with the variable."""
    if not isinstance(value, variable.type):
        raise ValueError(f"{variable.name} has type {variable.type}, got {value!r}")

    if variable.choices and value not in variable.choices:
        raise ValueError(
            f"{variable.name} has choices {variable.choices}, got {value!r}"
        )



"""Command-line interface for creating projects from Cookiecutter templates."""
import pathlib
from typing import Iterator
from typing import Optional

import click

from cutty.entrypoints.cli._main import main as _main
from cutty.services.create import create as service_create


def extra_context_callback(
    context: click.Context, parameter: click.Parameter, args: tuple[str, ...]
) -> dict[str, str]:
    """Callback for the EXTRA_CONTEXT argument."""

    def _generate() -> Iterator[tuple[str, str]]:
        for arg in args:
            try:
                key, value = arg.split("=", 1)
                yield key, value
            except ValueError:
                raise click.BadParameter(
                    "EXTRA_CONTEXT should contain items of the form key=value; "
                    f"'{arg}' doesn't match that form"
                )

    return dict(_generate())


@_main.command()
@click.argument("template")
@click.argument("extra-context", nargs=-1, callback=extra_context_callback)
@click.option(
    "--no-input",
    is_flag=True,
    default=False,
    help="Do not prompt for template variables.",
)
@click.option(
    "-c",
    "--checkout",
    metavar="REV",
    help="Branch, tag, or commit hash of the template repository.",
)
@click.option(
    "-o",
    "--output-dir",
    metavar="DIR",
    type=click.Path(file_okay=False, dir_okay=True, path_type=pathlib.Path),
    help="Parent directory of the generated project.",
)
@click.option(
    "--directory",
    metavar="DIR",
    type=click.Path(path_type=pathlib.Path),
    help=(
        "Directory within the template repository that contains the "
        "cookiecutter.json file."
    ),
)
@click.option(
    "-f",
    "--overwrite-if-exists",
    is_flag=True,
    default=False,
    help="Overwrite the contents of the output directory if it already exists.",
)
@click.option(
    "-s",
    "--skip-if-file-exists",
    is_flag=True,
    default=False,
    help="Skip the files in the corresponding directories if they already exist.",
)
@click.option(
    "-i",
    "--in-place",
    is_flag=True,
    default=False,
    help="Strip the leading path component from generated files.",
)
def create(
    template: str,
    extra_context: dict[str, str],
    no_input: bool,
    checkout: Optional[str],
    output_dir: Optional[pathlib.Path],
    directory: Optional[pathlib.Path],
    overwrite_if_exists: bool,
    skip_if_file_exists: bool,
    in_place: bool,
) -> None:
    """cutty."""
    service_create(
        template,
        extra_context=extra_context,
        no_input=no_input,
        checkout=checkout,
        outputdir=output_dir,
        directory=pathlib.PurePosixPath(directory) if directory is not None else None,
        overwrite_if_exists=overwrite_if_exists,
        skip_if_file_exists=skip_if_file_exists,
        outputdirisproject=in_place,
    )



"""Filesystem path."""
from __future__ import annotations

from collections.abc import Iterator

from cutty.filesystems.domain.filesystem import Access
from cutty.filesystems.domain.filesystem import Filesystem
from cutty.filesystems.domain.purepath import PurePath
from cutty.util.typeguard_ignore import typeguard_ignore


class Path(PurePath):
    """Location in a filesystem, with methods for accessing the filesystem."""

    filesystem: Filesystem

    def __init__(self, *parts: str, filesystem: Filesystem) -> None:
        """Initialize."""
        super().__init__(*parts)
        object.__setattr__(self, "filesystem", filesystem)

    def _withparts(self, *parts: str) -> Path:
        """Create a path with the given parts."""
        return Path(*parts, filesystem=self.filesystem)

    def is_dir(self) -> bool:
        """Return True if this is a directory."""
        return self.filesystem.is_dir(self)

    def iterdir(self) -> Iterator[Path]:
        """Iterate over the files in this directory."""
        for entry in self.filesystem.iterdir(self):
            yield self / entry

    def is_file(self) -> bool:
        """Return True if this is a regular file (or a symlink to one)."""
        return self.filesystem.is_file(self)

    def read_bytes(self) -> bytes:
        """Return the contents of this file."""
        return self.filesystem.read_bytes(self)

    def read_text(self) -> str:
        """Return the contents of this file."""
        return self.filesystem.read_text(self)

    def is_symlink(self) -> bool:
        """Return True if this is a symbolic link."""
        return self.filesystem.is_symlink(self)

    def readlink(self) -> PurePath:
        """Return the target of a symbolic link."""
        return self.filesystem.readlink(self)

    def access(self, mode: Access = Access.DEFAULT) -> bool:
        """Return True if the user can access the path."""
        return self.filesystem.access(self, mode)

    def __hash__(self) -> int:
        """Return the hash value of the path."""
        return hash((self.parts, self.filesystem))

    @typeguard_ignore
    def __eq__(self, other: object) -> bool:
        """Return True if the paths are equal."""
        if not isinstance(other, Path):
            return NotImplemented
        return self.filesystem is other.filesystem and self.filesystem.eq(self, other)

    @typeguard_ignore
    def __lt__(self, other: object) -> bool:
        """Return True if the path is less than the other path."""
        if not isinstance(other, Path):
            return NotImplemented
        if self.filesystem is not other.filesystem:
            raise ValueError("cannot compare paths on different filesystems")
        return self.filesystem.lt(self, other)

    @typeguard_ignore
    def __gt__(self, other: object) -> bool:
        """Return True if the path is greater than the other path."""
        if not isinstance(other, Path):
            return NotImplemented
        if self.filesystem is not other.filesystem:
            raise ValueError("cannot compare paths on different filesystems")
        return self.filesystem.lt(other, self)

    @typeguard_ignore
    def __le__(self, other: object) -> bool:
        """Return True if the path is less than or equal to the other path."""
        if not isinstance(other, Path):
            return NotImplemented
        if self.filesystem is not other.filesystem:
            raise ValueError("cannot compare paths on different filesystems")
        return not self.filesystem.lt(other, self)

    @typeguard_ignore
    def __ge__(self, other: object) -> bool:
        """Return True if the path is greater than or equal to the other path."""
        if not isinstance(other, Path):
            return NotImplemented
        if self.filesystem is not other.filesystem:
            raise ValueError("cannot compare paths on different filesystems")
        return not self.filesystem.lt(self, other)



"""Provider for Mercurial repositories."""
from cutty.repositories.adapters.fetchers.mercurial import hgfetcher
from cutty.repositories.domain.providers import remoteproviderfactory


hgproviderfactory = remoteproviderfactory(fetch=[hgfetcher])
